import logging

import pika
from pika.channel import Channel

from popug_sdk.amqp.utils import get_connection_params
from popug_sdk.conf import settings
from popug_sdk.conf.amqp import AMQPSettings

logger = logging.getLogger(settings.project)


class BaseProducer:
    def __init__(self, config: AMQPSettings, init_exchange: bool = False):
        self._config = config
        self._init_exchange = init_exchange

        self._connection: pika.BlockingConnection | None = None
        self._channel: Channel | None = None

    @property
    def is_closed_connection(self) -> bool:
        return not self._connection or self._connection.is_closed

    @property
    def is_closed_channel(self) -> bool:
        return not self._channel or self._channel.is_closed

    def _open_connection(self) -> None:
        if self.is_closed_connection or self.is_closed_channel:

            if self.is_closed_connection:
                parameters = get_connection_params(self._config)
                logger.info(
                    f"Connecting to host={parameters.host} "
                    f"port={parameters.port} "
                    f"virtual_host={parameters.virtual_host}>"
                )
                self._connection = pika.BlockingConnection(parameters)

            if self.is_closed_channel:
                self.channel = self._connection.channel()  # type: ignore

            if self._init_exchange is True:
                logger.info(
                    f"Creating a new exchange. "
                    f"Exchange: {self._config.exchange.name}"
                )
                self._channel.exchange_declare(  # type: ignore
                    exchange=self._config.exchange.name,
                    exchange_type=self._config.exchange.type.value,
                    durable=self._config.exchange.durable,
                )
                self._init_exchange = False

    def publish_message(self, body: bytes, routing_key: str) -> None:
        if self.is_closed_connection or self.is_closed_channel:
            self._open_connection()

        logger.info(
            f"Publishing to routing_key {routing_key}, "
            f"exchange {self._config.exchange.name}, message {body!r}"
        )

        self.channel.basic_publish(
            exchange=self._config.exchange.name,
            routing_key=routing_key,
            body=body,
            properties=pika.BasicProperties(
                content_type="application/json", delivery_mode=2
            ),
        )



from pydantic import BaseModel

from popug_sdk.conf.constants import (
    BASE_NAME,
    LOCALHOST,
    PortType,
)


class DatabaseSettings(BaseModel):
    enabled: bool = False
    host: str = LOCALHOST
    port: PortType = 5432
    user: str = "postgres"
    password: str = "postgres"
    database_name: str = f"{BASE_NAME}_db"



from typing import Callable

from redis import Redis

from popug_sdk.redis.redis_pool import get_redis_pool


def get_redis(config_name: str = "default") -> Callable[[], Redis]:
    def _get_redis() -> Redis:
        return Redis(connection_pool=get_redis_pool(config_name))

    return _get_redis



from __future__ import annotations

import importlib
import os
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Type,
)

__all__ = ("settings",)

if TYPE_CHECKING:
    from popug_sdk.conf.global_settings import Settings as BaseSettings


class Settings:
    ENVIRONMENT_VARIABLE = "SETTINGS_MODULE"

    @classmethod
    @lru_cache
    def load(cls) -> BaseSettings:
        settings_module = os.environ.get(
            cls.ENVIRONMENT_VARIABLE, "popug_sdk.conf.global_settings"
        )
        module = importlib.import_module(settings_module)
        settings_class: Type[BaseSettings] = getattr(module, "Settings")
        _settings = settings_class()
        _settings.set_up_logging()

        return _settings


settings = Settings.load()



from contextlib import contextmanager
from typing import Any

from sqlalchemy import create_engine as _engine_create
from sqlalchemy.engine import Engine
from sqlalchemy.orm import (
    Session,
    sessionmaker,
)

from popug_sdk.conf import settings
from popug_sdk.db.registry import mapper_registry  # noqa

__all__ = (
    "create_engine",
    "create_session",
    "init_db",
    "mapper_registry",
)


_engine: Engine | None = None
_Session: sessionmaker | None = None


def create_engine() -> Engine:
    global _engine

    if not _engine:
        _engine = _engine_create(settings.database_dsn, pool_pre_ping=True)

    return _engine


def init_db() -> sessionmaker:
    global _Session
    create_engine()

    if not _Session:
        _Session = sessionmaker(
            autocommit=False,
            autoflush=False,
            bind=_engine,
            expire_on_commit=False,
        )

        return _Session


@contextmanager
def create_session(session: Session | None = None, **kwargs: Any) -> Session:
    if not session:
        _session = init_db()

        with _session(**kwargs) as session:
            yield session
    else:
        yield session



import uuid
from datetime import datetime

from pydantic import (
    BaseModel,
    Field,
)


class BaseEvent(BaseModel):
    pid: str = Field(default_factory=lambda: str(uuid.uuid4()))
    version: int = 1
    created_at: datetime = Field(default_factory=datetime.utcnow)



import os
import pkgutil
from importlib import import_module
from typing import (
    Callable,
    NoReturn,
)

import typer

from popug_sdk.conf import settings

init_commands = typer.Typer()


def find_commands(management_dir: str) -> list[str]:
    command_dir = os.path.join(management_dir, "commands")
    return [
        name
        for _, name, is_pkg in pkgutil.iter_modules([command_dir])
        if not is_pkg and not name.startswith("_")
    ]


def get_commands() -> dict[str, str]:
    commands = {name: "popug_sdk" for name in find_commands(__path__[0])}
    for config_path in settings.commands_roots:
        path = os.path.join(config_path, "management")
        commands.update({name: config_path for name in find_commands(path)})

    return commands


def get_command_fail(exc: Exception, name: str) -> Callable[[], NoReturn]:
    def _command_fail() -> NoReturn:
        """Failed to load command"""
        raise exc

    _command_fail.__name__ = name

    return _command_fail


for command_name, base_path in get_commands().items():
    try:
        module = import_module(
            f"{base_path}.management.commands.{command_name}"
        )
    except ModuleNotFoundError as exc:
        command_function = get_command_fail(exc, command_name)
    else:
        command_function = getattr(module, command_name)

    help_text = command_function.__doc__
    init_commands.command(help=help_text)(command_function)



from datetime import datetime
from typing import Any

from fastapi import (
    APIRouter,
    FastAPI,
)
from fastapi.middleware.httpsredirect import HTTPSRedirectMiddleware
from fastapi.openapi.utils import get_openapi

from popug_sdk.conf import settings
from popug_sdk.schemas.pong import Pong

__all__ = ("app",)


def custom_openapi(app: FastAPI) -> dict[str, Any]:
    if app.openapi_schema:
        return app.openapi_schema

    openapi_schema = get_openapi(
        title=settings.project,
        version=settings.version or "0.0.1",
        routes=app.routes,
    )
    openapi_schema["x-tagGroups"] = settings.tag_groups
    app.openapi_schema = openapi_schema

    return app.openapi_schema


healthcheck_router = APIRouter()


@healthcheck_router.get("/", response_model=Pong)
def ping() -> dict[str, Any]:
    """Healthcheck endpoint"""

    return {
        "project": settings.project,
        "version": settings.version,
        "datetime": datetime.utcnow().isoformat(),
    }


app = FastAPI(title=settings.project)


if settings.use_https:
    app.add_middleware(HTTPSRedirectMiddleware)

app.openapi = lambda: custom_openapi(app)  # type: ignore
app.include_router(healthcheck_router, prefix="")


@app.on_event("startup")
def startup() -> None:
    if settings.database.enabled:
        from popug_sdk.db import init_db  # noqa

        init_db()

    if settings.redis:
        from popug_sdk.redis.redis_pool import init_redis_pool  # noqa

        init_redis_pool(settings.redis)


@app.on_event("shutdown")
def shutdown() -> None:
    if settings.redis:
        from popug_sdk.redis.redis_pool import close_redis_pool  # noqa

        close_redis_pool(settings.redis.keys())
